""" Unit tests for modern price queries

Copyright (c) 2021-, Tellor Development Community
Distributed under the terms of the MIT License.
"""
from telliot_core.queries.coin_price import CoinPrice


def test_constructor():
    """Validate price query"""
    q = CoinPrice(coin="BTC", currency="USD")

    # exp = b'["CoinPrice",{"coin":"BTC","currency":"USD","price_type":"current"}]'
    exp = b'{"type":"CoinPrice","coin":"BTC","currency":"USD","price_type":"current"}'

    print(q.query_data)
    assert q.query_data == exp

    exp = "c6bc01753088c82583c6e011d736314e4548c05052908cd63b4207b78fbc6f40"
    assert q.query_id.hex() == exp


def test_price_type():
    """Validate price_type setting"""
    q = CoinPrice(coin="ETH", currency="USD", price_type="24hr_twap")

    exp = b'{"type":"CoinPrice","coin":"ETH","currency":"USD","price_type":"24hr_twap"}'

    print(q.query_data)
    assert q.query_data == exp

    exp = "055367425f0ea7d0f53cdbe7f782ae348d2c721a26734da5f3c612f090e6262e"
    assert q.query_id.hex() == exp



"""  Oracle Query Module

"""
import json

from clamfig import Serializable
from telliot_core.types.value_type import ValueType
from web3 import Web3


class OracleQuery(Serializable):
    """Oracle Query

    An OracleQuery specifies how to pose a question to the
    Tellor Oracle and how to format/interpret the response.

    The OracleQuery class serves
    as the base class for all Queries, and implements default behaviors.
    Each subclass corresponds to a unique Query Type supported
    by the TellorX network.

    All public attributes of an OracleQuery represent an input that can
    be used to customize the query.

    The base class provides:

    - Calculation of the contents of the `data` field to include with the
      `TellorX.Oracle.tipQuery()` contract call.

    - Calculation of the `id` field field to include with the
      `TellorX.Oracle.tipQuery()` and `TellorX.Oracle.submitValue()`
      contract calls.

    """

    @property
    def descriptor(self) -> str:
        """Get the query descriptor string.
        The Query descriptor is a unique string representation of the query.
        The descriptor is required for users to specify the query to TellorX
        through the ``TellorX.Oracle.tipQuery()`` contract call.


        """
        state = self.get_state()
        jstr = json.dumps(state, separators=(",", ":"))
        return jstr

    @property
    def value_type(self) -> ValueType:
        """Returns the ValueType expected by the current Query configuration

        The value type defines required data type/structure of the
        ``value`` submitted to the contract through
        ``TellorX.Oracle.submitValue()``

        This method must be overridden by subclasses
        """
        pass

    @property
    def query_data(self) -> bytes:
        """Returns the ``data`` field for use in ``TellorX.Oracle.tipQuery()``
        contract call.

        """
        return self.descriptor.encode("utf-8")

    @property
    def query_id(self) -> bytes:
        """Returns the query ``id`` for use with the
        ``TellorX.Oracle.tipQuery()`` and ``TellorX.Oracle.submitValue()``
        contract calls.
        """
        return bytes(Web3.keccak(self.query_data))



""" telliot_core.datafeed.data_source

"""
import random
from collections import deque
from dataclasses import dataclass
from dataclasses import field
from typing import Deque
from typing import Generic
from typing import List
from typing import TypeVar

from telliot_core.model.base import Base
from telliot_core.types.datapoint import DataPoint
from telliot_core.types.datapoint import datetime_now_utc
from telliot_core.types.datapoint import OptionalDataPoint

T = TypeVar("T")


@dataclass
class DataSource(Generic[T], Base):
    """Base Class for a DataSource

    A DataSource provides an input to a `DataFeed`
    It also contains a store for all previously fetched data points.

    All subclasses must implement `DataSource.fetch_new_datapoint()`
    """

    max_datapoints: int = 256

    # Private storage for fetched values
    _history: Deque[DataPoint[T]] = field(default_factory=deque, init=False, repr=False)

    def __post_init__(self) -> None:
        # Overwrite default deque
        self._history = deque(maxlen=self.max_datapoints)

    @property
    def latest(self) -> OptionalDataPoint[T]:
        """Returns the most recent datapoint or none if history is empty"""
        if len(self._history) >= 1:
            return self._history[-1]
        else:
            return None, None

    def store_datapoint(self, datapoint: DataPoint[T]) -> None:
        """Store a datapoint"""
        v, t = datapoint
        if v is not None and t is not None:
            self._history.append(datapoint)

    def get_all_datapoints(self) -> List[DataPoint[T]]:
        """Get a list of all available data points"""
        return list(self._history)

    async def fetch_new_datapoint(self) -> OptionalDataPoint[T]:
        """Fetch new value and store it for later retrieval"""
        raise NotImplementedError

    @property
    def depth(self) -> int:
        return len(self._history)


@dataclass
class RandomSource(DataSource[float]):
    """A random data source

    Returns a random floating point number in the range [0.0, 1.0).
    """

    async def fetch_new_datapoint(self) -> DataPoint[float]:
        fetched_value = random.random()
        timestamp = datetime_now_utc()
        datapoint = (fetched_value, timestamp)

        self.store_datapoint(datapoint)

        return datapoint


# class ConstantSource(DataSource):
#     """A simple data source that fetches a constant value"""
#
#     #: Descriptive name
#     name: str = "Constant"
#
#     #: Constant value
#     constant_value: float
#
#     def __init__(self, value: float, **kwargs: Any):
#         super().__init__(constant_value=value, **kwargs)
#
#     async def update_value(self):
#         return self.value
#



"""Legacy Query Example """
from telliot_core.queries import LegacyRequest

q = LegacyRequest(legacy_id=1)
print(q.descriptor)
print(f"tipQuery data: 0x{q.query_data.hex()}")
print(f"tipQuery ID: 0x{q.query_id.hex()}")

value = 10000.1234567
print(f"submitValue (float): {value}")

encoded_bytes = q.value_type.encode(value)
print(f"submitValue (bytes): 0x{encoded_bytes.hex()}")

decoded_value = q.value_type.decode(encoded_bytes)
print(f"Decoded value (float): {decoded_value}")



""" telliot_core.datafeed.data_feed

"""
from dataclasses import dataclass
from typing import Generic
from typing import TypeVar

from telliot_core.datasource import DataSource
from telliot_core.model.base import Base
from telliot_core.queries.query import OracleQuery

T = TypeVar("T")


@dataclass
class DataFeed(Generic[T], Base):
    """Data feed providing query response

    A data feed contains a DataSource to fetch values in response to an `OracleQuery`.

    Attributes:
        query: The Query that this feed responds to
        source: Data source for feed
    """

    query: OracleQuery

    source: DataSource[T]


if __name__ == "__main__":
    # Example:
    from telliot_core.datasource import RandomSource
    from telliot_core.queries.legacy_query import LegacyRequest

    feed = DataFeed(source=RandomSource(), query=LegacyRequest(legacy_id=99))

    import yaml

    print(yaml.dump(feed.get_state(), sort_keys=False))



import json
from pathlib import Path

from telliot_core.directory._tellorx.temp_contracts import rinkeby_tellor_master
from telliot_core.directory._tellorx.temp_contracts import rinkeby_tellor_oracle
from telliot_core.directory.base import ContractDirectory
from telliot_core.directory.base import ContractInfo

__all__ = ["tellor_directory"]

_tellor_address_mainnet = {
    "master": "0x88dF592F8eb5D7Bd38bFeF7dEb0fBc02cf3778a0",
    "controller": "",
    "oracle": "",
    "governance": "",
    "treasury": "",
}

_tellor_address_rinkeby = {
    "master": "0x88dF592F8eb5D7Bd38bFeF7dEb0fBc02cf3778a0",
    "controller": "0x0f2B0a8fa0f60459f51E452273C879eb32555e91",
    "oracle": "0x18431fd88adF138e8b979A7246eb58EA7126ea16",
    "governance": "0xA64Bb0078eB80c97484f3f09Adb47b9B73CBcA00",
    "treasury": "0x2dB91443f2b562B8b2B2e8E4fC0A3EDD6c195147",
}

# Read contract ABIs from json files
_abi_folder = Path(__file__).resolve().parent / "_tellorx"
_abi_dict = {}
for name in ["master", "controller", "oracle", "governance", "treasury"]:
    with open(_abi_folder / f"{name}_abi.json", "r") as f:
        _abi_dict[name] = json.load(f)

# Create default tellor directory
tellor_directory = ContractDirectory()

for name in ["master", "controller", "oracle", "governance", "treasury"]:
    # Add maininet contract info entry
    tellor_directory.add_contract(
        ContractInfo(
            chain_id=1,
            org="tellor",
            name=name,
            address=_tellor_address_mainnet[name],
            abi=_abi_dict[name],
        )
    )

    # Add rinkeby contract info entry
    tellor_directory.add_contract(
        ContractInfo(
            chain_id=4,
            org="tellor",
            name=name,
            address=_tellor_address_rinkeby[name],
            abi=_abi_dict[name],
        )
    )


# TEMPORARY OVERRIDE ADDRESS and ABI WITH TEST VALUES

_tellor_address_rinkeby_temp = {
    "master": "0x657b95c228A5de81cdc3F85be7954072c08A6042",
    "oracle": "0x07b521108788C6fD79F471D603A2594576D47477",
}

tellor_master_rinkeby = tellor_directory.find(chain_id=4, name="master")[0]
tellor_master_rinkeby.address = _tellor_address_rinkeby_temp["master"]
tellor_master_rinkeby.abi = rinkeby_tellor_master

tellor_oracle_rinkeby = tellor_directory.find(chain_id=4, name="oracle")[0]
tellor_oracle_rinkeby.address = _tellor_address_rinkeby_temp["oracle"]
tellor_oracle_rinkeby.abi = rinkeby_tellor_oracle



"""Pytest Fixtures used for testing Pytelliot"""
import os

import pytest
from telliot_core.apps.telliot_config import TelliotConfig
from telliot_core.contract.contract import Contract
from telliot_core.directory.tellorx import tellor_directory


@pytest.fixture(scope="session", autouse=True)
def rinkeby_cfg():
    """Get rinkeby endpoint from config

    If environment variables are defined, they will override the values in config files
    """
    cfg = TelliotConfig()

    # Override configuration for rinkeby testnet
    cfg.main.chain_id = 4

    rinkeby_endpoint = cfg.get_endpoint()
    # assert rinkeby_endpoint.network == "rinkeby"

    # Optionally override private key and URL with ENV vars for testing
    if os.getenv("PRIVATE_KEY", None):
        cfg.main.private_key = os.environ["PRIVATE_KEY"]

    if os.getenv("NODE_URL", None):
        rinkeby_endpoint.url = os.environ["NODE_URL"]

    return cfg


@pytest.fixture(scope="session")
def master(rinkeby_cfg):
    """Helper function for connecting to a contract at an address"""
    tellor_master = tellor_directory.find(chain_id=4, name="master")[0]
    endpoint = rinkeby_cfg.get_endpoint()
    endpoint.connect()
    master = Contract(
        address=tellor_master.address,  # "0x657b95c228A5de81cdc3F85be7954072c08A6042",
        abi=tellor_master.abi,
        node=endpoint,
        private_key=rinkeby_cfg.main.private_key,
    )
    master.connect()
    return master


@pytest.fixture(scope="session", autouse=True)
def oracle(rinkeby_cfg):
    """Helper function for connecting to a contract at an address"""
    tellor_oracle = tellor_directory.find(chain_id=4, name="oracle")[0]
    endpoint = rinkeby_cfg.get_endpoint()
    endpoint.connect()
    oracle = Contract(
        address=tellor_oracle.address,  # "0x07b521108788C6fD79F471D603A2594576D47477",
        abi=tellor_oracle.abi,
        node=endpoint,
        private_key=rinkeby_cfg.main.private_key,
    )
    oracle.connect()
    return oracle



""" Unit tests for data source module

"""
from datetime import datetime

import pytest
from telliot_core.datasource import RandomSource


@pytest.mark.asyncio
async def test_random_source():
    """Test the random source example"""
    s = RandomSource()

    assert s.latest == (None, None)

    v, t = await s.fetch_new_datapoint()
    assert isinstance(v, float)
    assert 0 <= v < 1
    assert isinstance(t, datetime)

    v, t = await s.fetch_new_datapoint()

    assert s.depth == 2

    latest_values = s.get_all_datapoints()
    assert len(latest_values) == 2



import asyncio
import logging

from telliot_core.contract.listener import AddressEventListener
from telliot_core.contract.listener import event_message_handler
from telliot_core.contract.listener import listener_client
from telliot_core.directory.tellorx import tellor_directory

logger = logging.getLogger("__name__")


async def tellor_listener_client(ws_url: str, chain_id: int) -> None:
    master_info = tellor_directory.find(name="master", chain_id=chain_id)[0]
    oracle_info = tellor_directory.find(name="oracle", chain_id=chain_id)[0]

    listeners = [
        AddressEventListener(
            address=master_info.address, handler=event_message_handler
        ),
        AddressEventListener(
            address=oracle_info.address, handler=event_message_handler
        ),
    ]

    await listener_client(ws_url=ws_url, listeners=listeners)


if __name__ == "__main__":

    logging.basicConfig(level=logging.INFO)
    logger.info("begin test")
    from telliot_core.apps.telliot_config import TelliotConfig

    rinkeby_cfg = TelliotConfig()
    rinkeby_cfg.main.chain_id = 4

    ep = rinkeby_cfg.get_endpoint()
    if ep:
        url = ep.url

    async def run_listener() -> None:

        task = asyncio.create_task(tellor_listener_client(ws_url=url, chain_id=4))

        # Wait for 1 second
        await asyncio.sleep(120)

        task.cancel()

        try:
            logger.info("awaiting task")
            await task
        except asyncio.CancelledError:
            logger.info("run_listener(): Task cancelled")

    asyncio.run(run_listener())



""" Telliot application helpers

"""
import logging
import threading
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import Optional

from telliot_core.apps.telliot_config import TelliotConfig
from telliot_core.utils.home import telliot_homedir

logger = logging.getLogger(__name__)


@dataclass
class Application:
    """Application base class"""

    #: Application Name
    name: str

    #: Home directory
    homedir: Path = field(default_factory=telliot_homedir)

    #: Application configuration object
    config: Optional[TelliotConfig] = None

    #: Private thread storage
    _thread: Optional[threading.Thread] = None
    _shutdown: threading.Event = field(default_factory=threading.Event)

    def __post_init__(self) -> None:

        if not self.config:
            self.config = TelliotConfig(config_dir=self.homedir)

        # Logging
        self.configure_logging()

        logger.info("Created new {} application object".format(self.name))
        logger.info("Home Directory: {}".format(self.homedir))

    def configure_logging(self) -> None:
        """Configure Application logging

        Subclasses may override this method as required
        """

        # type checking does not seem to recognize that config
        # and homedir were validated/coerced in __init__
        assert self.config is not None
        assert isinstance(self.homedir, Path)

        # Convert loglevel text to log type
        loglevel = eval("logging." + self.config.main.loglevel.upper())

        logfile = self.homedir / (self.name + ".log")
        logging.basicConfig(filename=str(logfile), level=loglevel)

    def startup(self) -> None:
        """Startup Application

        Start the main application thread
        """
        logger.info("Starting {} Application".format(self.name))
        self._thread = threading.Thread(target=self.main, name=self.name)
        self._thread.start()

    def shutdown(self) -> None:
        """Startup Application

        Send a shutdown event to the main application thread and wait
        for it to terminate.
        """
        if self._thread:
            logger.info("Terminating {} Application".format(self.name))

            self._shutdown.set()

            #: Join worker thread
            self._thread.join()

            self._thread = None

    def main(self) -> None:
        """Main worker thread

        This method defined the main application processing.
        The main thread must monitor and respond to the self._shutdown event.

        This method *must* be overridded by subclasses.
        """
        seconds = 0
        while not self._shutdown.wait(1.0):
            print("Main thread processing: {}".format(seconds))
            seconds += 1

        print("Application {} received shutdown event".format(self.name))
        self._shutdown.clear()



from telliot_core.directory.tellorx import tellor_directory


def test_telliot_directory():

    tellor_contracts = tellor_directory.find(org="tellor")
    assert len(tellor_contracts) == 10

    rinkeby_contracts = tellor_directory.find(org="tellor", chain_id=4)
    assert len(rinkeby_contracts) == 5

    tellor_master_rinkeby = tellor_directory.find(
        org="tellor", name="master", chain_id=4
    )
    assert len(tellor_master_rinkeby) == 1



""" :mod:`telliot_core.queries.price_query`

"""
from dataclasses import dataclass
from typing import Literal

from telliot_core.queries.query import OracleQuery
from telliot_core.types.float_type import UnsignedFloatType
from telliot_core.types.value_type import ValueType

price_types = Literal["current", "eod", "24hr_twap", "1hr_twap"]

# List of inputs used to customize a CoinPrice object
price_query_params = ["coin", "currency", "price_type"]


@dataclass
class CoinPrice(OracleQuery):
    """Query the price of a cryptocurrency coin. **PROTOTYPE: DO NOT USE**

    Attributes:
        coin: Token symbol
        currency: Price currency symbol (default = USD)
        price_type: Price Type (default = current)

    """

    coin: str = ""

    currency: str = "usd"

    price_type: price_types = "current"

    @property
    def value_type(self) -> ValueType:
        """Returns the fixed value type for a CoinPrice."""
        return UnsignedFloatType(abi_type="ufixed64x6", packed=True)


if __name__ == "__main__":
    """CoinPrice Example."""

    q = CoinPrice(coin="btc")
    print(q.__repr__())

    print(q.query_data)
    print("0x" + q.query_id.hex())



from dataclasses import dataclass
from typing import Optional

from telliot_core.model.base import Base


@dataclass
class ResponseStatus(Base):
    ok: bool = True
    error: Optional[str] = None
    e: Optional[Exception] = None



""" :mod:`telliot_core.queries.value_type`

"""
from dataclasses import dataclass
from typing import Any

from clamfig import Serializable
from eth_abi.abi import decode_single
from eth_abi.abi import encode_single
from eth_abi.packed import encode_single_packed


@dataclass
class ValueType(Serializable):
    """Value Type

    A ValueType specifies the data structure of ``value`` included in
    the ``TellorX.Oracle.submitValue()`` used in response to
    tip request.

    The type is specified per eth-abi grammar, i.e.

    - https://eth-abi.readthedocs.io/en/latest/grammar.html
    """

    # ABI Encoding type string
    abi_type: str = "uint256"

    #: True if the value should be encoded using packed bytes format.
    packed: bool = False

    def encode(self, value: Any) -> bytes:
        """Encode a value using the ABI Type string."""
        if self.packed:
            return encode_single_packed(self.abi_type, value)
        else:
            return encode_single(self.abi_type, value)

    def decode(self, bytes_val: bytes) -> Any:
        """Decode bytes into a value using abi type string."""
        return decode_single(self.abi_type, bytes_val)

    # @validator("abi_type")
    # def require_valid_grammar(cls, v: str) -> str:
    #     """A validator to require well formed ABI type string grammar."""
    #     t = eth_abi.grammar.parse(v)
    #     t.validate()
    #     return eth_abi.grammar.normalize(v)  # type: ignore
    #
    # def json(self, **kwargs: Any) -> str:
    #     """Return compact json format used in query descriptor"""
    #
    #     return super().json(**kwargs, separators=(",", ":"))



from dataclasses import dataclass
from dataclasses import field

from telliot_core.datasource import DataSource
from telliot_core.pricing.price_service import WebPriceService
from telliot_core.types.datapoint import OptionalDataPoint


@dataclass
class PriceSource(DataSource[float]):
    """Current Asset Price

    The Current Asset Price data source retrieves the price of a coin
    in the specified current from a `WebPriceService`.
    """

    #: Asset symbol
    asset: str = ""

    #: Price currency symbol
    currency: str = ""

    #: Price Service
    service: WebPriceService = field(default_factory=WebPriceService)  # type: ignore

    async def fetch_new_datapoint(self) -> OptionalDataPoint[float]:
        """Update current value with time-stamped value fetched from source

        Returns:
            New datapoint
        """
        datapoint = await self.service.get_price(self.asset, self.currency)
        v, t = datapoint
        if v is not None and t is not None:
            self.store_datapoint((v, t))

        return datapoint



from dataclasses import dataclass

from telliot_core.queries.query import OracleQuery


def test_main():
    @dataclass
    class MyQuery(OracleQuery):
        text: str
        val: int = 3

    q = MyQuery("asdf")
    state = q.get_state()
    print(state)
    assert state == {"type": "MyQuery", "text": "asdf", "val": 3}



"""Test plugin module"""
import pytest
from telliot_core.plugin.discover import telliot_plugins


@pytest.mark.skip("Fix test after plugin move complete")
def test_discovered_plugins():
    # Make sure that default telliot_examples plugin package is registered
    assert "telliot_feed_examples" in telliot_plugins

    example_plugin = telliot_plugins["telliot_feed_examples"]

    _ = example_plugin.registry


# def test_plugin_registry():
#     """Test barebones plugin registry interface"""
#
#     class MyQueryType(OracleQuery):
#         pass
#
#
#     myfeed = MyFeedType(query=MyQueryType())
#
#     r = PluginRegistry()
#
#     r.register_query_type(MyQueryType)
#     r.register_feed_type(MyFeedType)
#
#     r.register_feed(myfeed)
#
#     assert myfeed in r.feeds
#     assert MyQueryType in r.query_types
#     assert MyFeedType in r.feed_types



""" Unit tests for static queries

Copyright (c) 2021-, Tellor Development Community
Distributed under the terms of the MIT License.
"""
from telliot_core.queries.string_query import StringQuery


def test_static_query():
    """Test static query"""
    q = StringQuery(text="What is the meaning of life")

    assert (
        q.query_data == b'{"type":"StringQuery","text":"What is the meaning of life"}'
    )

    submit_value = q.value_type.encode(
        "Please refer to: https://en.wikipedia.org/wiki/Meaning_of_life"
    )
    assert isinstance(submit_value, bytes)



from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import Literal
from typing import Optional
from typing import Union

from telliot_core.apps.config import ConfigFile
from telliot_core.apps.config import ConfigOptions
from telliot_core.model.base import Base
from telliot_core.model.chain import ChainList
from telliot_core.model.endpoints import EndpointList
from telliot_core.model.endpoints import RPCEndpoint


@dataclass
class MainConfig(ConfigOptions):
    """Main telliot_core configuration object"""

    #: Control application logging level
    loglevel: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] = "INFO"

    #: Select chain id
    chain_id: int = 4

    #: Select network (e.g. mainnet, testnet, rinkeby
    network: str = "rinkeby"

    #: Private key for selected chain_id/network
    private_key: str = ""


@dataclass
class TelliotConfig(Base):
    """Main telliot_core configuration object

    Load all configuration objects from config files.
    If any config file does not exist, a default will be created.
    """

    config_dir: Optional[Union[str, Path]] = None

    main: MainConfig = field(default_factory=MainConfig)

    endpoints: EndpointList = field(default_factory=EndpointList)

    chains: ChainList = field(default_factory=ChainList)

    def __post_init__(self) -> None:
        main_file = ConfigFile(
            name="main",
            config_type=MainConfig,
            config_format="yaml",
            config_dir=self.config_dir,
        )
        ep_file = ConfigFile(
            name="endpoints",
            config_type=EndpointList,
            config_format="yaml",
            config_dir=self.config_dir,
        )
        chain_file = ConfigFile(
            name="chains",
            config_type=ChainList,
            config_format="json",
            config_dir=self.config_dir,
        )

        self.main = main_file.get_config()
        self.endpoints = ep_file.get_config()
        self.chains = chain_file.get_config()

    def get_endpoint(self) -> Optional[RPCEndpoint]:
        """Search endpoints for current chain_id"""
        return self.endpoints.get_chain_endpoint(self.main.chain_id)



import json
from typing import Literal

import requests

ethgastypes = Literal["fast", "fastest", "safeLow", "average"]


async def fetch_gas_price() -> int:
    """Estimate current ETH gas price

    Current implementation fetches from ethgasstation

    Returns:
        eth gas price in gwei
    """
    return await ethgasstation("fast")


async def ethgasstation(style: ethgastypes = "fast") -> int:
    """Fetch gas price from ethgasstation in gwei"""
    rsp = requests.get("https://ethgasstation.info/json/ethgasAPI.json")
    prices = json.loads(rsp.content)
    gas_price = int(prices[style])

    return int(gas_price / 10)  # json output is gwei*10



""" Listener Module

This module adds contract event listening capability for websocket endpoints
(like infura) that is lacking from web3.py.

The goal is to allow users to subscribe to and define async callbacks for each event.

*Work in Progress*
"""
import asyncio
import logging
from dataclasses import dataclass
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import List

import aiohttp
from web3._utils.method_formatters import log_entry_formatter
from web3.types import LogReceipt

logger = logging.getLogger("__name__")

AsyncCallable = Callable[[Any], Awaitable]


async def event_message_handler(msg: Any) -> None:
    log_receipt: LogReceipt = log_entry_formatter(msg)
    logger.info(f"handled event message: {log_receipt}")


async def subscribe_contract_events(
    ws: aiohttp.ClientWebSocketResponse, contract_address: str
) -> None:
    logger.info(f"Subscribing to contract events at address {contract_address}")
    request = {
        "jsonrpc": "2.0",
        "id": 1,
        "method": "eth_subscribe",
        "params": ["logs", {"address": contract_address}],
    }

    await ws.send_json(request)

    subscription_response = await ws.receive_json()
    _ = subscription_response.get("id")
    sub_result = subscription_response.get("result")
    logger.info(
        f"Subscription complete: result={sub_result} for address {contract_address}"
    )


async def subscribe_blocks(ws: aiohttp.ClientWebSocketResponse) -> None:
    logger.info("Subscribing to newHeads")

    msg = {"jsonrpc": "2.0", "id": 1, "method": "eth_subscribe", "params": ["newHeads"]}

    await ws.send_json(msg)

    subscription_response = await ws.receive_json()
    _ = subscription_response.get("id")
    sub_result = subscription_response.get("result")

    logger.info(f"newHead subscription complete: result={sub_result}")


async def receive_messages(
    ws: aiohttp.ClientWebSocketResponse, handler: AsyncCallable
) -> None:
    while True:
        try:
            message = await asyncio.wait_for(ws.receive_json(), timeout=60)
            await handler(message)
        except asyncio.CancelledError:
            logger.debug("Listener cancelled")
            raise
        except asyncio.exceptions.TimeoutError:
            logger.debug("Listener Timed out, restarting.")
            continue


async def block_listener(session: aiohttp.ClientSession, ws_url: str) -> None:
    async def block_message_handler(msg: Any) -> None:
        sub_id = msg["params"]["subscription"]
        logger.info(f"handled block subscription: {sub_id}")

    async with session.ws_connect(ws_url) as ws2:
        await subscribe_blocks(ws2)
        await receive_messages(ws2, block_message_handler)


async def contract_event_listener(
    session: aiohttp.ClientSession,
    ws_url: str,
    address: str,
    message_handler: AsyncCallable,
) -> None:
    async with session.ws_connect(ws_url) as ws1:
        await subscribe_contract_events(ws1, address)
        await receive_messages(ws1, message_handler)


@dataclass
class AddressEventListener:
    """Address Event Listener

    A listener that handles all event logs from a contract address

    Attributes:
        address: Contract address
        handler: Function that takes the received JSON message as a single argument

    """

    address: str
    handler: AsyncCallable


async def listener_client(ws_url: str, listeners: List[AddressEventListener]) -> None:
    """Top level Listener Client

    This function creates a single session that is shared by multiple listeners.

    Args:
        listeners: List of listeners to add to the main client session
        ws_url: URL of the websocket to use for each listener connection
    """

    async with aiohttp.ClientSession() as session:
        logger.info("Listener session created")

        tasks = [block_listener(session, ws_url)]
        for listener in listeners:
            tasks.append(
                contract_event_listener(
                    session, ws_url, listener.address, listener.handler
                )
            )

        await asyncio.gather(*tasks)



""" :mod:`telliot_core.queries.legacy_query`

"""
from dataclasses import dataclass

from telliot_core.queries.query import OracleQuery
from telliot_core.types.float_type import UnsignedFloatType
from telliot_core.types.value_type import ValueType


@dataclass
class LegacyRequest(OracleQuery):
    """Legacy Price/Value request

    Legacy request are queries that existed prior to TellorX
    A legacy query uses arbitrary query ``data`` and a static query ``id``.
    The query ``id`` is always set to the legacy request ID, which is
    an integer less than 100.

    The LegacyQuery class is deprecated and should not be used by
    new projects.  Instead, use the
    [`CoinPrice`][telliot_core.queries.coin_price.CoinPrice]
    query or create a new query.

    Refer to [tellor documentation](https://docs.tellor.io/tellor/integration/data-ids)
    for a description of each ``id``

    """

    legacy_id: int
    """The request ID of all legacy queries is a static integer 1 < N <=100"""

    @property
    def value_type(self) -> ValueType:
        """Returns the Legacy Value Type for all legacy queries"""
        return UnsignedFloatType(abi_type="ufixed256x6", packed=False)

    @property
    def query_id(self) -> bytes:
        """Override query ``id`` with the legacy request ID."""
        return self.legacy_id.to_bytes(32, "big", signed=False)



""" Telliot CLI

A simple interface for interacting with telliot_core's functionality.
Configure telliot_core's settings via this interface's command line flags
or in the configuration file.
"""
import click
import yaml
from telliot_core.apps.telliot_config import TelliotConfig


@click.group()
def main() -> None:
    """Telliot command line interface"""
    pass


@main.group()
def config() -> None:
    """Manage telliot_core configuration"""
    pass


@config.command()
def init() -> None:
    """Create initial configuration files"""
    _ = TelliotConfig()


@config.command()
def show() -> None:
    """Create initial configuration files"""
    cfg = TelliotConfig()
    print(yaml.dump(cfg.get_state(), sort_keys=False))


if __name__ == "__main__":
    main()



""" Simple example of creating a "plug-in" data feed

"""
from telliot_core.datafeed import DataFeed
from telliot_core.datasource import RandomSource
from telliot_core.queries.legacy_query import LegacyRequest


def test_data_feed():
    """Test a new feed"""
    myfeed = DataFeed(source=RandomSource(), query=LegacyRequest(legacy_id=4))

    import yaml

    print(yaml.dump(myfeed.get_state(), sort_keys=False))



""" :mod:`telliot_core.queries.string_query`

"""
# Copyright (c) 2021-, Tellor Development Community
# Distributed under the terms of the MIT License.
from dataclasses import dataclass

from telliot_core.queries.query import OracleQuery
from telliot_core.types.value_type import ValueType


@dataclass
class StringQuery(OracleQuery):
    """Static Oracle Query

    A text query supports a question in the form of an arbitrary
    text.
    """

    #: Static query text
    text: str

    @property
    def value_type(self) -> ValueType:
        """Returns a default text response type."""
        return ValueType(abi_type="string", packed=False)



""" :mod:`telliot_core.types`

The Types package provides mechanisms to specify response types
for Oracle Queries.
"""
# Copyright (c) 2021-, Tellor Development Community
# Distributed under the terms of the MIT License.



"""Text Query Example """
from telliot_core.queries import StringQuery

q = StringQuery(text="What is the meaning of life?")
print(q.descriptor)
print(f"tipQuery data: 0x{q.query_data.hex()}")
print(f"tipQuery ID: 0x{q.query_id.hex()}")

value = "Please refer to: https://en.wikipedia.org/wiki/Meaning_of_life"
print(f"submitValue (str): {value}")

encoded_bytes = q.value_type.encode(value)
print(f"submitValue (bytes): 0x{encoded_bytes.hex()}")

decoded_value = q.value_type.decode(encoded_bytes)
print(f"Decoded value (float): {decoded_value}")



rinkeby_tellor_master = [
    {
        "inputs": [{"internalType": "bytes32", "name": "_queryId", "type": "bytes32"}],
        "name": "getTimestampCountById",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "_governance", "type": "address"},
            {"internalType": "address", "name": "_oracle", "type": "address"},
            {"internalType": "address", "name": "_treasury", "type": "address"},
        ],
        "stateMutability": "nonpayable",
        "type": "constructor",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "_owner",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "_spender",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "_value",
                "type": "uint256",
            },
        ],
        "name": "Approval",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "address",
                "name": "_staker",
                "type": "address",
            }
        ],
        "name": "NewStaker",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "address",
                "name": "_staker",
                "type": "address",
            }
        ],
        "name": "StakeWithdrawRequested",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "address",
                "name": "_staker",
                "type": "address",
            }
        ],
        "name": "StakeWithdrawn",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "_from",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "_to",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "_value",
                "type": "uint256",
            },
        ],
        "name": "Transfer",
        "type": "event",
    },
    {"stateMutability": "nonpayable", "type": "fallback"},
    {
        "inputs": [
            {"internalType": "address", "name": "", "type": "address"},
            {"internalType": "address", "name": "", "type": "address"},
        ],
        "name": "_allowances",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "bytes", "name": "_b", "type": "bytes"}],
        "name": "_sliceUint",
        "outputs": [{"internalType": "uint256", "name": "_x", "type": "uint256"}],
        "stateMutability": "pure",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
        "name": "addresses",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "_user", "type": "address"},
            {"internalType": "address", "name": "_spender", "type": "address"},
        ],
        "name": "allowance",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "_user", "type": "address"},
            {"internalType": "uint256", "name": "_amount", "type": "uint256"},
        ],
        "name": "allowedToTrade",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "_spender", "type": "address"},
            {"internalType": "uint256", "name": "_amount", "type": "uint256"},
        ],
        "name": "approve",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "_from", "type": "address"},
            {"internalType": "address", "name": "_to", "type": "address"},
            {"internalType": "uint256", "name": "_amount", "type": "uint256"},
        ],
        "name": "approveAndTransferFrom",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "_user", "type": "address"}],
        "name": "balanceOf",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "_user", "type": "address"},
            {"internalType": "uint256", "name": "_blockNumber", "type": "uint256"},
        ],
        "name": "balanceOfAt",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "", "type": "address"},
            {"internalType": "uint256", "name": "", "type": "uint256"},
        ],
        "name": "balances",
        "outputs": [
            {"internalType": "uint128", "name": "fromBlock", "type": "uint128"},
            {"internalType": "uint128", "name": "value", "type": "uint128"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "uint256", "name": "_amount", "type": "uint256"}],
        "name": "burn",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
        "name": "bytesVars",
        "outputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "_newController", "type": "address"}
        ],
        "name": "changeControllerContract",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "_newGovernance", "type": "address"}
        ],
        "name": "changeGovernanceContract",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "_newOracle", "type": "address"}
        ],
        "name": "changeOracleContract",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "_reporter", "type": "address"},
            {"internalType": "uint256", "name": "_status", "type": "uint256"},
        ],
        "name": "changeStakingStatus",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "_newTreasury", "type": "address"}
        ],
        "name": "changeTreasuryContract",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes32", "name": "_target", "type": "bytes32"},
            {"internalType": "uint256", "name": "_amount", "type": "uint256"},
        ],
        "name": "changeUint",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "name": "currentMiners",
        "outputs": [
            {"internalType": "uint256", "name": "value", "type": "uint256"},
            {"internalType": "address", "name": "miner", "type": "address"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "decimals",
        "outputs": [{"internalType": "uint8", "name": "", "type": "uint8"}],
        "stateMutability": "pure",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "depositStake",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
        "name": "disputeIdByDisputeHash",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "name": "disputesById",
        "outputs": [
            {"internalType": "bytes32", "name": "hash", "type": "bytes32"},
            {"internalType": "int256", "name": "tally", "type": "int256"},
            {"internalType": "bool", "name": "executed", "type": "bool"},
            {"internalType": "bool", "name": "disputeVotePassed", "type": "bool"},
            {"internalType": "bool", "name": "isPropFork", "type": "bool"},
            {"internalType": "address", "name": "reportedMiner", "type": "address"},
            {"internalType": "address", "name": "reportingParty", "type": "address"},
            {
                "internalType": "address",
                "name": "proposedForkAddress",
                "type": "address",
            },
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "bytes32", "name": "_data", "type": "bytes32"}],
        "name": "getAddressVars",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "_disputeId", "type": "uint256"}
        ],
        "name": "getAllDisputeVars",
        "outputs": [
            {"internalType": "bytes32", "name": "", "type": "bytes32"},
            {"internalType": "bool", "name": "", "type": "bool"},
            {"internalType": "bool", "name": "", "type": "bool"},
            {"internalType": "bool", "name": "", "type": "bool"},
            {"internalType": "address", "name": "", "type": "address"},
            {"internalType": "address", "name": "", "type": "address"},
            {"internalType": "address", "name": "", "type": "address"},
            {"internalType": "uint256[9]", "name": "", "type": "uint256[9]"},
            {"internalType": "int256", "name": "", "type": "int256"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "bytes32", "name": "_hash", "type": "bytes32"}],
        "name": "getDisputeIdByDisputeHash",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "_disputeId", "type": "uint256"},
            {"internalType": "bytes32", "name": "_data", "type": "bytes32"},
        ],
        "name": "getDisputeUintVars",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "_requestId", "type": "uint256"}
        ],
        "name": "getLastNewValueById",
        "outputs": [
            {"internalType": "uint256", "name": "", "type": "uint256"},
            {"internalType": "bool", "name": "", "type": "bool"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "getNewCurrentVariables",
        "outputs": [
            {"internalType": "bytes32", "name": "_c", "type": "bytes32"},
            {"internalType": "uint256[5]", "name": "_r", "type": "uint256[5]"},
            {"internalType": "uint256", "name": "_diff", "type": "uint256"},
            {"internalType": "uint256", "name": "_tip", "type": "uint256"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "bytes32", "name": "_queryId", "type": "bytes32"}],
        "name": "getNewValueCountbyQueryId",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "_requestId", "type": "uint256"}
        ],
        "name": "getNewValueCountbyRequestId",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "_staker", "type": "address"}],
        "name": "getStakerInfo",
        "outputs": [
            {"internalType": "uint256", "name": "", "type": "uint256"},
            {"internalType": "uint256", "name": "", "type": "uint256"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes32", "name": "_queryId", "type": "bytes32"},
            {"internalType": "uint256", "name": "_index", "type": "uint256"},
        ],
        "name": "getTimestampbyQueryIdandIndex",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "_requestId", "type": "uint256"},
            {"internalType": "uint256", "name": "_index", "type": "uint256"},
        ],
        "name": "getTimestampbyRequestIDandIndex",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "bytes32", "name": "_data", "type": "bytes32"}],
        "name": "getUintVar",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "init",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "_addy", "type": "address"}],
        "name": "isMigrated",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "migrate",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "", "type": "address"}],
        "name": "migrated",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes32", "name": "", "type": "bytes32"},
            {"internalType": "address", "name": "", "type": "address"},
        ],
        "name": "minersByChallenge",
        "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "_receiver", "type": "address"},
            {"internalType": "uint256", "name": "_amount", "type": "uint256"},
        ],
        "name": "mint",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "name",
        "outputs": [{"internalType": "string", "name": "", "type": "string"}],
        "stateMutability": "pure",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "name": "newValueTimestamps",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
        "name": "requestIdByQueryHash",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "requestStakingWithdraw",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "_requestId", "type": "uint256"},
            {"internalType": "uint256", "name": "_timestamp", "type": "uint256"},
        ],
        "name": "retrieveData",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes32", "name": "_queryId", "type": "bytes32"},
            {"internalType": "uint256", "name": "_timestamp", "type": "uint256"},
        ],
        "name": "retrieveData",
        "outputs": [{"internalType": "bytes", "name": "", "type": "bytes"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "_reporter", "type": "address"},
            {"internalType": "address", "name": "_disputer", "type": "address"},
        ],
        "name": "slashReporter",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "symbol",
        "outputs": [{"internalType": "string", "name": "", "type": "string"}],
        "stateMutability": "pure",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "totalSupply",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "_to", "type": "address"},
            {"internalType": "uint256", "name": "_amount", "type": "uint256"},
        ],
        "name": "transfer",
        "outputs": [{"internalType": "bool", "name": "success", "type": "bool"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "_from", "type": "address"},
            {"internalType": "address", "name": "_to", "type": "address"},
            {"internalType": "uint256", "name": "_amount", "type": "uint256"},
        ],
        "name": "transferFrom",
        "outputs": [{"internalType": "bool", "name": "success", "type": "bool"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
        "name": "uints",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "verify",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "pure",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "withdrawStake",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes32", "name": "_queryId", "type": "bytes32"},
            {"internalType": "bytes", "name": "_value", "type": "bytes"},
            {"internalType": "uint256", "name": "_nonce", "type": "uint256"},
            {"internalType": "bytes", "name": "_queryData", "type": "bytes"},
        ],
        "name": "submitValue",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
]

rinkeby_tellor_oracle = [
    {
        "inputs": [{"internalType": "address", "name": "_master", "type": "address"}],
        "stateMutability": "nonpayable",
        "type": "constructor",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "bytes32",
                "name": "_queryId",
                "type": "bytes32",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "_time",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "bytes",
                "name": "_value",
                "type": "bytes",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "_reward",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "_nonce",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "bytes",
                "name": "_queryData",
                "type": "bytes",
            },
        ],
        "name": "NewReport",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "_newReportingLock",
                "type": "uint256",
            }
        ],
        "name": "ReportingLockChanged",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "_newTimeBasedReward",
                "type": "uint256",
            }
        ],
        "name": "TimeBasedRewardsChanged",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "_user",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "bytes32",
                "name": "_queryId",
                "type": "bytes32",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "_tip",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "_totalTip",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "bytes",
                "name": "_queryData",
                "type": "bytes",
            },
        ],
        "name": "TipAdded",
        "type": "event",
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "_newReportingLock", "type": "uint256"}
        ],
        "name": "changeReportingLock",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "depositStake",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {
                "internalType": "uint256",
                "name": "_newTimeBasedReward",
                "type": "uint256",
            }
        ],
        "name": "changeTimeBasedReward",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes32", "name": "_queryId", "type": "bytes32"},
            {"internalType": "uint256", "name": "_timestamp", "type": "uint256"},
        ],
        "name": "getBlockNumberByTimestamp",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "bytes32", "name": "_queryId", "type": "bytes32"}],
        "name": "getCurrentReward",
        "outputs": [
            {"internalType": "uint256", "name": "", "type": "uint256"},
            {"internalType": "uint256", "name": "", "type": "uint256"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "bytes32", "name": "_queryId", "type": "bytes32"}],
        "name": "getCurrentValue",
        "outputs": [{"internalType": "bytes", "name": "", "type": "bytes"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes32", "name": "_queryId", "type": "bytes32"},
            {"internalType": "uint256", "name": "_index", "type": "uint256"},
        ],
        "name": "getReportTimestampByIndex",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes32", "name": "_queryId", "type": "bytes32"},
            {"internalType": "uint256", "name": "_timestamp", "type": "uint256"},
        ],
        "name": "getReporterByTimestamp",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "_reporter", "type": "address"}],
        "name": "getReporterLastTimestamp",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "getReportingLock",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "_reporter", "type": "address"}],
        "name": "getReportsSubmittedByAddress",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "getTimeBasedReward",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "getTimeOfLastNewValue",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "bytes32", "name": "_queryId", "type": "bytes32"}],
        "name": "getTimestampCountById",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes32", "name": "_queryId", "type": "bytes32"},
            {"internalType": "uint256", "name": "_timestamp", "type": "uint256"},
        ],
        "name": "getTimestampIndexByTimestamp",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "bytes32", "name": "_queryId", "type": "bytes32"}],
        "name": "getTipsById",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "_user", "type": "address"}],
        "name": "getTipsByUser",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes32", "name": "_queryId", "type": "bytes32"},
            {"internalType": "uint256", "name": "_timestamp", "type": "uint256"},
        ],
        "name": "getValueByTimestamp",
        "outputs": [{"internalType": "bytes", "name": "", "type": "bytes"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes32", "name": "_queryId", "type": "bytes32"},
            {"internalType": "uint256", "name": "_timestamp", "type": "uint256"},
        ],
        "name": "removeValue",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "reportingLock",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes32", "name": "_queryId", "type": "bytes32"},
            {"internalType": "bytes", "name": "_value", "type": "bytes"},
            {"internalType": "uint256", "name": "_nonce", "type": "uint256"},
            {"internalType": "bytes", "name": "_queryData", "type": "bytes"},
        ],
        "name": "submitValue",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "timeBasedReward",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "timeOfLastNewValue",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes32", "name": "_queryId", "type": "bytes32"},
            {"internalType": "uint256", "name": "_tip", "type": "uint256"},
            {"internalType": "bytes", "name": "_queryData", "type": "bytes"},
        ],
        "name": "tipQuery",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
        "name": "tips",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "tipsInContract",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "verify",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "pure",
        "type": "function",
    },
]



from datetime import datetime
from datetime import timezone
from typing import Optional
from typing import Tuple
from typing import TypeVar


T = TypeVar("T")

# A Time-stamped value tuple
DataPoint = Tuple[T, datetime]

# An optional time-stamped value tuple
OptionalDataPoint = Tuple[Optional[T], Optional[datetime]]


def datetime_now_utc() -> datetime:
    """A helper function to get the timestamp for "now" """
    return datetime.now(timezone.utc)



# Copyright (c) 2021-, Tellor Development Community
# Copyright (c) 2020, Nucleic Development Team.
import os

import pytest

example_folder = os.path.join(os.path.dirname(__file__), "..", "examples")
examples = list()
for (dir_path, _, filenames) in os.walk(example_folder):
    examples += [os.path.join(dir_path, f) for f in filenames]


@pytest.mark.parametrize("path", examples)
def test_example(path):
    """Run all examples"""
    with open(path, "r") as f:
        # If exec gets two separate objects as globals and locals, the code
        # will be executed as if it were embedded in a class definition, which
        # changes name look up rules
        exec(f.read(), locals(), locals())



from dataclasses import dataclass
from dataclasses import field
from typing import Any
from typing import List
from typing import Type

from telliot_core.datafeed import DataFeed
from telliot_core.datasource import DataSource
from telliot_core.queries import OracleQuery


@dataclass
class PluginRegistry:
    """Plugin Registry

    This is the main interface for plugins to register capabilities with Telliot
    """

    # #: List of data feed instances to register with telliot_core
    # feeds: List[DataFeed] = field(default_factory=list)

    #: List of custom query types to register with telliot_core
    query_types: List[Type[OracleQuery]] = field(default_factory=list)

    #: List of custom data feed types to register with telliot_core
    feed_types: List[Type[DataFeed[Any]]] = field(default_factory=list)

    #: List of custom data feed types to register with telliot_core
    source_types: List[Type[DataSource[Any]]] = field(default_factory=list)

    # def register_feed(self, feed: DataFeed) -> None:
    #     """Register a feed"""
    #     self.feeds.append(feed)

    def register_query_type(self, query_type: Type[OracleQuery]) -> None:
        """Register a query_type"""
        self.query_types.append(query_type)

    def register_feed_type(self, feed_type: Type[DataFeed[Any]]) -> None:
        """Register a feed type"""
        self.feed_types.append(feed_type)

    def register_source_type(self, feed_type: Type[DataSource[Any]]) -> None:
        """Register a source type"""
        self.source_types.append(feed_type)



from dataclasses import dataclass
from dataclasses import field
from typing import Any
from typing import List
from typing import Optional


@dataclass
class ContractInfo:
    org: str
    name: str
    chain_id: int
    address: str
    abi: Optional[List[Any]] = field(repr=False)


@dataclass
class ContractDirectory:
    # Private directory storage
    _contents: List[ContractInfo] = field(default_factory=list)

    def add_contract(self, info: ContractInfo) -> None:
        self._contents.append(info)

    def find(
        self,
        *,
        org: str = "tellor",
        name: Optional[str] = None,
        address: Optional[str] = None,
        chain_id: Optional[int] = None,
    ) -> List[ContractInfo]:
        result = []
        for info in self._contents:
            if info.org != org:
                continue
            if chain_id is not None:
                if chain_id != info.chain_id:
                    continue
            if name is not None:
                if info.name is not name:
                    continue
            if address is not None:
                if info.address is not address:
                    continue

            result.append(info)

        return result



""" Unit tests for response type

Copyright (c) 2021-, Tellor Development Community
Distributed under the terms of the MIT License.
"""
from decimal import Decimal

import pytest
from eth_abi.exceptions import InsufficientDataBytes
from telliot_core.types.float_type import UnsignedFloatType
from telliot_core.types.value_type import ValueType


def test_fixed_response_type():
    """Demonstrate encoding a fixed value with precision=9"""
    value = Decimal("1.0")
    r1 = ValueType(abi_type="ufixed256x9", packed=False)
    bytes_val = r1.encode(value)
    assert (
        bytes_val.hex()
        == "000000000000000000000000000000000000000000000000000000003b9aca00"
    )
    int_val = int.from_bytes(bytes_val, "big", signed=False)
    assert int_val == 10 ** 9


def test_dynamic_response_type_unpacked():
    """Demonstrate a complex response to a query"""
    r1 = ValueType(abi_type="(int8,bytes,ufixed32x9,bool[])[2]", packed=False)

    value = ((1, b"abc", 1, (True, True)), (1, b"def", 1, (True, True)))

    bytes_val = r1.encode(value)
    print(bytes_val.hex())

    decoded = r1.decode(bytes_val)

    assert decoded == value


def test_packed_response_type_FAILS():
    """Demonstrate encoding a fixed value with precision=9

    This test demonstrates that a custom decoder is required for
    packed values
    """
    value = Decimal("1.0")
    r1 = ValueType(abi_type="ufixed64x9", packed=True)
    bytes_val = r1.encode(value)
    assert bytes_val.hex() == "000000003b9aca00"
    int_val = int.from_bytes(bytes_val, "big", signed=False)
    assert int_val == 10 ** 9

    with pytest.raises(InsufficientDataBytes):
        decoded = r1.decode(bytes_val)
        print(decoded)


def test_unsigned_float_value():
    f = UnsignedFloatType(abi_type="ufixed64x6")
    assert not f.packed
    assert f.decimals == 6
    # Note encoded value is still 256 bits because packed = False
    assert f.nbits == 64
    print(f.get_state())
    encoded_value = f.encode(99.0000009)
    print(encoded_value.hex())
    assert (
        encoded_value.hex()
        == "0000000000000000000000000000000000000000000000000000000005e69ec1"
    )

    decoded_value = f.decode(encoded_value)
    assert isinstance(decoded_value, float)
    assert decoded_value == 99.000001



"""
Unit tests covering telliot_core CLI commands.
"""
from click.testing import CliRunner
from telliot_core.cli import main


def test_config_cmd():
    """Test telliot_core CLI command: report."""
    runner = CliRunner()
    result = runner.invoke(main, ["config", "init"])

    print(result)



"""
Utils for creating a JSON RPC connection to an EVM blockchain
"""
from dataclasses import dataclass
from dataclasses import field
from typing import List
from typing import Optional

from telliot_core.apps.config import ConfigFile
from telliot_core.apps.config import ConfigOptions
from telliot_core.model.base import Base
from web3 import Web3
from web3.middleware import geth_poa_middleware


@dataclass
class RPCEndpoint(Base):
    """JSON RPC Endpoint for EVM compatible network"""

    #: Chain ID
    chain_id: Optional[int] = None

    #: Network Name (e.g. 'mainnet', 'testnet', 'rinkeby')
    network: str = ""

    #: Provider Name (e.g. 'Infura')
    provider: str = ""

    #: URL (e.g. 'https://mainnet.infura.io/v3/<project_id>')
    url: str = ""

    #: Explorer URL ')
    explorer: Optional[str] = None

    #: Read-only Web3 Connection with private storage
    web3 = property(lambda self: self._web3)
    _web3: Optional[Web3] = field(default=None, init=False, repr=False)

    def connect(self) -> bool:
        """Connect to EVM blockchain

        returns:
            True if connection was successful
        """

        if self._web3:
            return True

        if self.url.startswith("ws"):
            self._web3 = Web3(Web3.WebsocketProvider(self.url))
        elif self.url.startswith("http"):
            self._web3 = Web3(Web3.HTTPProvider(self.url))
        else:
            raise ValueError(f"Invalid endpoint url: {self.url}")

        # Inject middlware if connecting to rinkeby
        try:
            if self.network == "rinkeby":
                self.web3.middleware_onion.inject(geth_poa_middleware, layer=0)
        except Exception:
            print("unable to connect to rinkeby")

        try:
            connected = self._web3.isConnected()
        # Pokt nodes won't submit isConnected rpc call
        except Exception:
            connected = self._web3.eth.get_block_number() > 1
        if connected:
            print("Connected to {}".format(self))
        else:
            print("Could not connect to {}".format(self))

        return connected


default_endpoint_list = [
    RPCEndpoint(
        chain_id=1,
        provider="Infura",
        network="mainnet",
        url="wss://mainnet.infura.io/ws/v3/{INFURA_API_KEY}",
        explorer="https://etherscan.io",
    ),
    RPCEndpoint(
        chain_id=4,
        provider="Infura",
        network="rinkeby",
        url="wss://rinkeby.infura.io/ws/v3{INFURA_API_KEY}",
        explorer="https://rinkeby.etherscan.io",
    ),
    RPCEndpoint(
        chain_id=137,
        provider="Matic",
        network="mainnet",
        url="https://rpc-mainnet.matic.network",
        explorer="https://matic.network",
    ),
]


@dataclass
class EndpointList(ConfigOptions):
    endpoints: List[RPCEndpoint] = field(default_factory=lambda: default_endpoint_list)

    def get_chain_endpoint(self, chain_id: int = 1) -> Optional[RPCEndpoint]:
        """Get an Endpoint for the specified chain_id"""

        for endpoint in self.endpoints:
            if endpoint.chain_id == chain_id:
                return endpoint

        return None


if __name__ == "__main__":
    cf = ConfigFile(name="endpoints", config_type=EndpointList, config_format="yaml")

    config_endpoints = cf.get_config()

    print(config_endpoints)



""" :mod:`telliot_core.types.float_type`

"""
from dataclasses import dataclass
from decimal import Decimal
from typing import Any

from telliot_core.types.value_type import ValueType


@dataclass
class UnsignedFloatType(ValueType):
    """Unsigned Float Type

    This class specifies the a floating point value
    using an ABI data type.  It also provides encoding/decoding
    to/from floating point values.

    """

    #: ABI Encoding for Unsigned Float value (default = ufixed256x6)
    abi_type: str = "ufixed256x6"

    @property
    def decimals(self) -> int:
        """Get precision from abi type"""
        mxn = self.abi_type[6:]
        m, n = mxn.split("x")
        return int(n)

    @property
    def nbits(self) -> int:
        """Get number of bits from abi type"""
        mxn = self.abi_type[6:]
        m, n = mxn.split("x")
        return int(m)

    def encode(self, value: float) -> bytes:
        """An encoder for float values

        This encoder converts a float value to the CoinPrice ABI
        data type.
        """

        decimal_value = Decimal(value).quantize(Decimal(10) ** -self.decimals)

        return super().encode(decimal_value)

    def decode(self, bytes_val: bytes) -> Any:
        """A decoder for float values

        This decoder converts from the CoinPrice ABI data type to
        a floating point value.
        """
        nbytes = self.nbits / 8

        if self.packed:
            if len(bytes_val) != nbytes:
                raise ValueError(f"Value must be {nbytes} bytes")

        intval = int.from_bytes(bytes_val, "big", signed=False)

        return intval / 10.0 ** self.decimals



""" Test application utilities

"""
import pathlib
import shutil
import threading
import time
from pathlib import Path

from telliot_core.apps.app import Application
from telliot_core.utils.home import default_homedir

# Temp file location
testhome = Path("./temp").resolve().absolute()


def test_homedir():
    """Test default home directory"""
    hd = default_homedir()
    assert isinstance(hd, pathlib.Path)
    assert hd.exists()


def test_application_homedir():
    """Test home directory handling"""
    if testhome.exists():
        shutil.rmtree(testhome)

    testhome.mkdir(parents=True)
    app = Application(name="appname", homedir=testhome)
    assert isinstance(app.homedir, pathlib.Path)

    shutil.rmtree(testhome)


def test_application_default_home():
    """Test default application directory"""

    app = Application(name="appname")
    assert app.homedir == default_homedir()
    assert "telliot" in str(app.homedir)


def test_application_subclassing():
    """Test application features"""

    # Create Application subclass
    # Note: other attributes-specific can be added as required
    class MyApp(Application):

        # Insert other attributes here

        def __init__(self, **data):
            super().__init__(name="myapp", **data)

    # Make sure temp dir exists
    testhome.mkdir(parents=True, exist_ok=True)

    #: Instantiate application subclass
    app = MyApp(homedir=testhome)
    assert app.homedir == testhome

    shutil.rmtree(testhome)


def test_application_processing():
    """Test application startup and shutdown"""
    testhome.mkdir(parents=True, exist_ok=True)

    app = Application(name="myapp", homedir=testhome)
    assert isinstance(app._shutdown, threading.Event)

    app = Application(name="myapp", homedir=testhome)
    app.startup()
    time.sleep(2)
    app.shutdown()

    shutil.rmtree(testhome)



"""telliot_core.apps.config module"""
import json
import logging
from pathlib import Path
from typing import Literal
from typing import Optional
from typing import Type
from typing import Union

import yaml
from clamfig import deserialize
from clamfig import Serializable
from clamfig import serialize
from telliot_core.utils.home import telliot_homedir
from yaml import CDumper as Dumper
from yaml import CLoader as Loader

logger = logging.getLogger(__name__)

config_formats = Literal["yaml", "json"]


class ConfigOptions(Serializable):
    """An object used to manage configuration options

    Each attribute represents a configuration option.
    Subclasses should add configuration attributes as required.
    """

    pass


class ConfigFile:
    """ConfigFile"""

    @property
    def config_file(self) -> Path:
        return self.config_dir / f"{self.name}.{self.config_format}"

    def __init__(
        self,
        name: str,
        config_type: Type[ConfigOptions],
        config_dir: Optional[Union[str, Path]] = None,
        config_format: config_formats = "yaml",
    ) -> None:
        """Construct a new ConfigFile object

        Args:
            name:
                Config filename

            config_dir:
                Config file folder  If None, home directory is automatically computed.

            config_format:
                Format of config file ('yaml' or 'json')

        """

        #: Configuration Name
        self.name = name

        #: Config Type
        self.config_type = config_type

        #: Configuration Folder
        self.config_dir = telliot_homedir(config_dir)

        #: File Format
        self.config_format = config_format

        # Try to load configuration from file
        if self.config_file.exists():

            # Try to load file
            _ = self.get_config()

        else:
            # Create a default configuration and save it
            options = self.config_type()
            self.save_config(options)

    def get_config(self) -> ConfigOptions:
        """Load Configuration from a .yaml file"""

        if self.config_format == "yaml":

            with open(self.config_file, "r") as f:

                state = yaml.load(f, Loader=Loader)

                # Try to parse file
                config = deserialize(state)

        elif self.config_format == "json":

            with open(self.config_file, "r") as f:
                state = json.load(f)
                config = deserialize(state)

        else:
            raise AttributeError(f"Invalid config file type: {self.config_format}")

        return config  # type: ignore

    def save_config(self, config: ConfigOptions) -> None:

        """Save configuration to file"""

        try:

            # Make sure folder exists
            self.config_file.parent.mkdir(parents=True, exist_ok=True)

            # Back up existing file
            if self.config_file.exists():
                self.config_file.rename(self.config_file.with_suffix(".bak"))

            if self.config_format == "yaml":

                with open(self.config_file, "w") as f:

                    state = serialize(config)
                    yaml.dump(state, f, Dumper=Dumper, sort_keys=False)

            elif self.config_format == "json":

                with open(self.config_file, "w") as f:
                    state = serialize(config)
                    jstr = json.dumps(state, indent=2)
                    f.write(jstr)

            print("Saved config '{}' to {}".format(self.name, self.config_file))

        except FileNotFoundError as e:
            print(
                "Error saving {} to {}".format(
                    self.__class__.__name__, self.config_file
                )
            )
            raise e



import asyncio
import logging

import pytest
from telliot_core.contract.tellor_listener import tellor_listener_client

logger = logging.getLogger(__name__)

logging.basicConfig(level=logging.INFO)


@pytest.mark.skip("Doesn't work in pytest")
def test_main(rinkeby_cfg, caplog):
    caplog.set_level(logging.INFO)

    url = rinkeby_cfg.get_endpoint().url

    async def run_listener() -> None:
        task = asyncio.create_task(tellor_listener_client(ws_url=url, chain_id=4))

        # Wait for 1 second
        await asyncio.sleep(120)

        task.cancel()

        try:
            logger.info("awaiting task")
            await task
        except asyncio.CancelledError:
            logger.info("run_listener(): Task cancelled")

    asyncio.run(run_listener())


@pytest.mark.skip("TODO")
def test_tellor_listener(rinkeby_cfg, event_loop, caplog):
    caplog.set_level(logging.INFO)
    print(rinkeby_cfg)

    from telliot_core.apps.telliot_config import TelliotConfig

    rinkeby_cfg = TelliotConfig()
    rinkeby_cfg.main.chain_id = 4

    url = rinkeby_cfg.get_endpoint().url

    async def run_listener():

        task = event_loop.create_task(tellor_listener_client(ws_url=url, chain_id=4))

        # Wait for 1 second
        await asyncio.sleep(30)

        task.cancel()

        try:
            logger.info("awaiting task")
            await task
        except asyncio.CancelledError:
            logger.info("run_listener(): Task cancelled")

    event_loop.run_until_complete(run_listener())



from abc import ABC
from abc import abstractmethod
from typing import Any
from typing import Dict

import requests
from telliot_core.types.datapoint import OptionalDataPoint


class PriceServiceInterface(ABC):
    """Price Service Interface

    Interface to get pricing information
    As an interface, this class stores no to_state and all methods are abstract.
    Classes that inherit this interface must provide concrete implementations
    of each method.
    """

    @abstractmethod
    async def get_price(self, asset: str, currency: str) -> OptionalDataPoint[float]:
        """Fetch the price of an coin

        TODO: Strictly specify compliant coin/currency symbols so concrete
            classes can comply.

        Args:
            asset: Asset (Ticker Symbol)
            currency: Currency of returned price (Ticker Symbol)

        Returns:
            Time-stamped coin price or None if an exception occurs
        """
        raise NotImplementedError


class WebPriceService(PriceServiceInterface, ABC):
    """Abstract Base CLass for a Web-based Pricing Service"""

    def __init__(self, name: str, url: str, timeout: float = 5.0):

        self.name = name
        self.url = url
        self.timeout = timeout

    def get_url(self, url: str = "") -> Dict[str, Any]:
        """Helper function to get URL JSON response while handling exceptions

        Args:
            url: URL to fetch

        Returns:
            A dictionary with the following (optional) keys:
                json (dict or list): Result, if no error occurred
                error (str): A description of the error, if one occurred
                exception (Exception): The exception, if one occurred
        """

        request_url = self.url + url

        with requests.Session() as s:
            try:
                r = s.get(request_url, timeout=self.timeout)
                json_data = r.json()
                return {"response": json_data}

            except requests.exceptions.ConnectTimeout as e:
                return {"error": "Timeout Error", "exception": e}

            except Exception as e:
                return {"error": str(type(e)), "exception": e}



"""
Utils for connecting to an EVM contract
"""
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

from eth_typing.evm import ChecksumAddress
from telliot_core.model.endpoints import RPCEndpoint
from telliot_core.utils.response import ResponseStatus
from web3 import Web3
from web3.datastructures import AttributeDict


class Contract:
    """Convenience wrapper for connecting to an Ethereum contract"""

    def __init__(
        self,
        address: Union[str, ChecksumAddress],
        abi: Union[List[Dict[str, Any]], str],
        node: RPCEndpoint,
        private_key: str = "",
    ):

        self.address = Web3.toChecksumAddress(address)
        self.abi = abi
        self.node = node
        self.contract = None
        self.private_key = private_key

    def connect(self) -> ResponseStatus:
        """Connect to EVM contract through an RPC Endpoint"""

        if not self.node.web3:
            msg = "node is not instantiated"
            return ResponseStatus(ok=False, error=msg)

        self.node.connect()
        self.contract = self.node.web3.eth.contract(address=self.address, abi=self.abi)
        return ResponseStatus(ok=True)

    async def read(
        self, func_name: str, **kwargs: Any
    ) -> Tuple[Optional[Tuple[Any]], ResponseStatus]:
        """
        Reads data from contract
        inputs:
        func_name (str): name of contract function to call

        returns:
        ResponseStatus: standard response for contract data
        """

        if self.contract:
            try:
                contract_function = self.contract.get_function_by_name(func_name)
                output = contract_function(**kwargs).call()
                return output, ResponseStatus(ok=True)
            except ValueError as e:
                msg = f"function '{func_name}' not found in contract abi"
                return None, ResponseStatus(ok=False, e=e, error=msg)
        else:
            msg = "no instance of contract"
            return None, ResponseStatus(ok=False, error=msg)

    async def write(
        self, func_name: str, gas_price: int, acc_nonce: int, **kwargs: Any
    ) -> Tuple[Optional[AttributeDict[Any, Any]], ResponseStatus]:
        """For submitting any contract transaction once without retries

        gas price measured in gwei

        """

        status = ResponseStatus()

        if not self.contract:
            msg = "unable to connect to contract"
            return None, ResponseStatus(ok=False, error=msg)

        if not self.node:
            msg = "no node instance"
            return None, ResponseStatus(ok=False, error=msg)

        if self.private_key:
            acc = self.node.web3.eth.account.from_key(self.private_key)
        else:
            msg = "Private key missing"
            return None, ResponseStatus(ok=False, error=msg)
        try:
            # build transaction
            contract_function = self.contract.get_function_by_name(func_name)
            transaction = contract_function(**kwargs)
            # estimated_gas = transaction.estimateGas()
            gas_limit = 500000  # TODO optimize for gas/profitability
            print("gas limit:", gas_limit)

            print("address: ----- ", acc.address)
            print("gas price:", gas_price)

            built_tx = transaction.buildTransaction(
                {
                    "from": acc.address,
                    "nonce": acc_nonce,
                    "gas": gas_limit,
                    "gasPrice": self.node.web3.toWei(gas_price, "gwei"),
                    "chainId": self.node.chain_id,
                }
            )

            # submit transaction
            tx_signed = acc.sign_transaction(built_tx)
            print(" tx signed")
            tx_hash = self.node.web3.eth.send_raw_transaction(tx_signed.rawTransaction)
            print("tx sent")
            # Confirm transaction
            tx_receipt = self.node.web3.eth.wait_for_transaction_receipt(
                tx_hash, timeout=360
            )

            # Point to relevant explorer
            print(
                f"""View reported data: \n
                {self.node.explorer}/tx/{tx_hash.hex()}
                """
            )

            return tx_receipt, status
        except Exception as e:
            status.ok = False
            status.error = str(e.args)
            status.e = e
            return None, status

    async def write_with_retry(
        self,
        func_name: str,
        gas_price: int,
        extra_gas_price: int,
        retries: int,
        **kwargs: Any,
    ) -> Tuple[Optional[AttributeDict[Any, Any]], ResponseStatus]:
        """For submitting any contract transaction. Retries supported!

        gas_price measured in gwei

        """

        try:
            status = ResponseStatus()
            acc = self.node.web3.eth.account.from_key(self.private_key)
            acc_nonce = self.node.web3.eth.get_transaction_count(acc.address)

            # Iterate through retry attempts
            for _ in range(retries + 1):

                tx_receipt, status = await self.write(
                    func_name=func_name,
                    gas_price=gas_price,
                    acc_nonce=acc_nonce,
                    **kwargs,
                )

                print("write status: ", status)

                # Exit loop if transaction successful
                if tx_receipt and status.ok and tx_receipt["status"] == 1:
                    print(
                        f"tx was successful! check it out at {self.node.explorer}/tx/{tx_receipt['transactionHash']}"  # noqa: E501
                    )  # noqa: E501
                    return tx_receipt, status
                elif (
                    not status.ok
                    and status.error
                    and "replacement transaction underpriced" in status.error
                ):
                    gas_price += extra_gas_price
                elif not status.ok and status.error and "already known" in status.error:
                    acc_nonce += 1
                elif not status.ok and status.error and "nonce too low" in status.error:
                    acc_nonce += 1
                # a different rpc error
                elif (
                    not status.ok
                    and status.error
                    and "nonce is too low" in status.error
                ):
                    acc_nonce += 1
                elif (
                    not status.ok
                    and status.error
                    and "not in the chain" in status.error
                ):
                    gas_price += extra_gas_price
                elif (
                    status.ok
                    and tx_receipt["status"] == 0  # type: ignore # error won't be none
                ):
                    status.error = "tx reverted by contract/evm logic"
                    print(f"tx was reverted by evm! check it out at {self.node.explorer}/tx/{tx_receipt['transactionHash']}")  # type: ignore # tx receipt won't be none # noqa: E501
                    return tx_receipt, status
                else:
                    extra_gas_price = 0

            status.ok = False
            status.error = "ran out of retries, tx unsuccessful"

            return tx_receipt, status

        except Exception as e:
            status.ok = False
            status.error = str(e.args)
            status.e = e
            return None, status

    def listen(self) -> None:
        """Wrapper for listening for contract events"""
        pass



""" Unit tests for Legacy Queries

Copyright (c) 2021-, Tellor Development Community
Distributed under the terms of the MIT License.
"""
from telliot_core.queries.legacy_query import LegacyRequest


def test_legacy_query():
    """Validate legacy query"""
    q = LegacyRequest(
        legacy_id=100,
    )
    assert q.value_type.abi_type == "ufixed256x6"
    assert q.value_type.packed is False

    exp = b'{"type":"LegacyRequest","legacy_id":100}'

    # print(q.query_data)
    assert q.query_data == exp

    assert (
        q.query_id.hex()
        == "0000000000000000000000000000000000000000000000000000000000000064"
    )



"""
Test covering Pytelliot EVM contract connection utils.
"""
import pytest
import web3
from hexbytes.main import HexBytes
from telliot_core.contract.gas import fetch_gas_price
from telliot_core.queries.legacy_query import LegacyRequest
from web3 import Web3


def test_connect_to_tellor(rinkeby_cfg, master):
    """Contract object should access Tellor functions"""
    assert len(master.contract.all_functions()) > 0
    assert isinstance(
        master.contract.all_functions()[0], web3.contract.ContractFunction
    )


@pytest.mark.skip(reason="for playground.")
@pytest.mark.asyncio
async def test_call_read_function(rinkeby_cfg, master):
    """Contract object should be able to call arbitrary contract read function"""

    output, status = await master.read(
        func_name="getTimestampCountById", _queryId=Web3.keccak(HexBytes(2))
    )
    assert status.ok
    assert output >= 0


@pytest.mark.skip(reason="oracle contract does not have faucet right now")
@pytest.mark.asyncio
async def test_faucet(rinkeby_cfg, master):
    """Contract call to mint to an account with the contract faucet"""
    # estimate gas
    gas_price = await fetch_gas_price()
    # set up user
    user = master.node.web3.eth.account.from_key(rinkeby_cfg.main.private_key).address
    # read balance
    balance1, status = await master.read(func_name="balanceOf", _user=user)
    assert status.ok
    assert balance1 >= 0
    print(balance1)
    # mint tokens to user
    receipt, status = await master.write_with_retry(
        func_name="setBalanceTest",
        gas_price=gas_price,
        extra_gas_price=20,
        retries=1,
        _address=user,
        _amount=1e18,
    )
    assert status.ok
    # read balance again
    balance2, status = await master.read(func_name="balanceOf", _user=user)
    assert status.ok
    print(balance2)
    assert balance2 - balance1 == 1e21


@pytest.mark.skip("Move to end-to-end tests")
@pytest.mark.asyncio
async def test_trb_transfer(rinkeby_cfg, master):
    """Test TRB transfer through TellorMaster contract (and its proxies)"""

    gas_price = await fetch_gas_price()
    sender = master.node.web3.eth.account.from_key(rinkeby_cfg.main.private_key).address
    balance, status = await master.read("balanceOf", _user=sender)
    print("my sender address:", sender)
    print("my sender balance:", balance / 1e18)
    recipient = str(
        Web3.toChecksumAddress("0xf3428C75CAfb3FBA46D3E190B7539Fbbfb96f244")
    )

    balance, status = await master.read("balanceOf", _user=recipient)
    assert status.ok
    print("before:", balance / 1e18)

    receipt, status = await master.write_with_retry(
        "transfer",
        _to=recipient,
        _amount=1,
        gas_price=gas_price,
        extra_gas_price=20,
        retries=2,
    )
    print(status.error)
    assert status.ok

    balance, status = await master.read("balanceOf", _user=recipient)
    print("after:", balance / 1e18)


@pytest.mark.skip("Move to end-to-end tests")
@pytest.mark.asyncio
async def test_submit_value(rinkeby_cfg, master, oracle):
    """E2E test for submitting a value to rinkeby"""

    gas_price_gwei = await fetch_gas_price()
    user = master.node.web3.eth.account.from_key(rinkeby_cfg.main.private_key).address
    print(user)

    balance, status = await master.read("balanceOf", _user=user)
    print(balance / 1e18)

    is_staked, status = await master.read("getStakerInfo", _staker=user)
    print(is_staked)

    if is_staked[0] == 0:
        _, status = await master.write_with_retry(
            func_name="depositStake",
            gas_price=gas_price_gwei,
            extra_gas_price=20,
            retries=2,
        )
        assert status.ok

    q = LegacyRequest(legacy_id=99)
    value = q.value_type.encode(420.0)

    query_data = q.query_data
    query_id = q.query_id

    timestamp_count, status = await oracle.read(
        func_name="getTimestampCountById", _queryId=query_id
    )
    assert status.ok
    assert timestamp_count >= 0
    print(timestamp_count)

    receipt, status = await oracle.write_with_retry(
        func_name="submitValue",
        gas_price=gas_price_gwei,
        extra_gas_price=40,
        retries=5,
        _queryId=query_id,
        _value=value,
        _nonce=timestamp_count,
        _queryData=query_data,
    )

    print(status.error)
    assert status.ok



""" telliot_core.queries

The Queries package provides a mechanism to define Oracle Query Types
and specify the format of the query and the response.
"""
# Copyright (c) 2021-, Tellor Development Community
# Distributed under the terms of the MIT License.
from telliot_core.queries.coin_price import CoinPrice
from telliot_core.queries.legacy_query import LegacyRequest
from telliot_core.queries.query import OracleQuery
from telliot_core.queries.string_query import StringQuery

__all__ = ["OracleQuery", "LegacyRequest", "StringQuery", "CoinPrice"]
